import logging
from typing import Any

import numpy as np
import numpy.typing as npt
import xarray as xr
from rasterio import warp

LOGGER = logging.getLogger(__name__)

# Earth-Centered, Earth-Fixed Coordinate Reference System used to express satellite orbit
# https://en.wikipedia.org/wiki/Earth-centered,_Earth-fixed_coordinate_system
# https://spatialreference.org/ref/epsg/wgs-84-2/
ECEF_CRS = "EPSG:4978"


def open_dem_raster(
    dem_urlpath: str, engine: str = "rasterio", **kwargs: Any
) -> xr.DataArray:
    dem_raster = xr.open_dataarray(dem_urlpath, engine=engine, **kwargs)
    if dem_raster.y.diff("y").values[0] < 0:
        dem_raster = dem_raster.isel(y=slice(None, None, -1))
    dem_raster.attrs["long_name"] = "elevation"
    dem_raster.attrs["units"] = "m"
    dem_raster = dem_raster.rename("dem").squeeze(drop=True)
    return dem_raster


def make_nd_dataarray(das: list[xr.DataArray], dim: str = "axis") -> xr.DataArray:
    da_nd = xr.concat(das, dim=dim, coords="minimal")
    return da_nd.assign_coords({dim: range(len(das))})


def convert_to_dem_3d(
    dem_raster: xr.DataArray, dim: str = "axis", x: str = "x", y: str = "y"
) -> xr.DataArray:
    _, dem_raster_x = xr.broadcast(dem_raster, dem_raster.coords[x])
    dem_raster_y = dem_raster.coords[y]
    dem_3d = make_nd_dataarray([dem_raster_x, dem_raster_y, dem_raster], dim=dim)
    return dem_3d.rename("dem_3d")


def transform_dem_3d(
    dem_3d: xr.DataArray,
    source_crs: str | None = None,
    target_crs: str = ECEF_CRS,
    dim: str = "axis",
) -> xr.DataArray:
    if source_crs is None:
        source_crs = dem_3d.rio.crs
    try:
        x, y, z = warp.transform(
            source_crs,
            target_crs,
            dem_3d.sel({dim: 0}).values.flat,
            dem_3d.sel({dim: 1}).values.flat,
            dem_3d.sel({dim: 2}).values.flat,
        )
    except Exception:
        # HACK: the very first call to warp.transform sometimes fails
        LOGGER.warn("rasterio.warp.transform failed, retrying...")
        x, y, z = warp.transform(
            source_crs,
            target_crs,
            dem_3d.sel({dim: 0}).values.flat,
            dem_3d.sel({dim: 1}).values.flat,
            dem_3d.sel({dim: 2}).values.flat,
        )
    dem_3d_crs: xr.DataArray = xr.zeros_like(dem_3d)
    shape = dem_3d_crs.loc[{dim: 0}].shape
    dem_3d_crs.loc[{dim: 0}] = np.reshape(x, shape)
    dem_3d_crs.loc[{dim: 1}] = np.reshape(y, shape)
    dem_3d_crs.loc[{dim: 2}] = np.reshape(z, shape)
    return dem_3d_crs


def convert_to_dem_ecef(
    dem_raster: xr.DataArray, x: str = "x", y: str = "y", **kwargs: Any
) -> xr.DataArray:
    dem_3d = convert_to_dem_3d(dem_raster, x=x, y=y)
    return transform_dem_3d(dem_3d, **kwargs)


def compute_dem_oriented_area(dem_ecef: xr.DataArray) -> xr.DataArray:
    x_corners: npt.ArrayLike = np.concatenate(
        [
            [dem_ecef.x[0] + (dem_ecef.x[0] - dem_ecef.x[1]) / 2],
            ((dem_ecef.x.shift(x=-1) + dem_ecef.x) / 2)[:-1].data,
            [dem_ecef.x[-1] + (dem_ecef.x[-1] - dem_ecef.x[-2]) / 2],
        ]
    )
    y_corners: npt.ArrayLike = np.concatenate(
        [
            [dem_ecef.y[0] + (dem_ecef.y[0] - dem_ecef.y[1]) / 2],
            ((dem_ecef.y.shift(y=-1) + dem_ecef.y) / 2)[:-1].data,
            [dem_ecef.y[-1] + (dem_ecef.y[-1] - dem_ecef.y[-2]) / 2],
        ]
    )
    dem_ecef_corners = dem_ecef.interp(
        {"x": x_corners, "y": y_corners},
        method="linear",
        kwargs={"fill_value": "extrapolate"},
    )

    dx = dem_ecef_corners.diff("x", 1)
    dy = dem_ecef_corners.diff("y", 1)

    dx1 = dx.isel(y=slice(1, None)).assign_coords(dem_ecef.coords)
    dy1 = dy.isel(x=slice(1, None)).assign_coords(dem_ecef.coords)
    dx2 = dx.isel(y=slice(None, -1)).assign_coords(dem_ecef.coords)
    dy2 = dy.isel(x=slice(None, -1)).assign_coords(dem_ecef.coords)

    cross_1 = xr.cross(dx1, dy1, dim="axis") / 2
    sign_1 = np.sign(
        xr.dot(cross_1, dem_ecef, dims="axis")
    )  # ensure direction out of DEM

    cross_2 = xr.cross(dx2, dy2, dim="axis") / 2
    sign_2 = np.sign(
        xr.dot(cross_2, dem_ecef, dims="axis")
    )  # ensure direction out of DEM
    dem_oriented_area: xr.DataArray = cross_1 * sign_1 + cross_2 * sign_2

    return dem_oriented_area.rename("dem_oriented_area")



from typing import Any, Optional, Tuple

import attrs
import numpy as np
import pandas as pd
import xarray as xr

S_TO_NS = 10**9


def polyder(coefficients: xr.DataArray) -> xr.DataArray:
    # TODO: raise if "degree" coord is not decreasing
    derivative_coefficients = coefficients.isel(degree=slice(1, None)).copy()
    for degree in coefficients.coords["degree"].values[:-1]:
        derivative_coefficients.loc[{"degree": degree - 1}] = (
            coefficients.loc[{"degree": degree}] * degree
        )
    return derivative_coefficients


@attrs.define
class OrbitPolyfitIterpolator:
    coefficients: xr.DataArray
    epoch: np.datetime64
    interval: Tuple[np.datetime64, np.datetime64]

    @classmethod
    def from_position(
        cls,
        position: xr.DataArray,
        dim: str = "azimuth_time",
        deg: int = 5,
        epoch: Optional[np.datetime64] = None,
        interval: Optional[Tuple[np.datetime64, np.datetime64]] = None,
    ) -> "OrbitPolyfitIterpolator":
        time = position.coords[dim]
        assert time.dtype.name in ("datetime64[ns]", "timedelta64[ns]")

        if epoch is None:
            # NOTE: summing two datetime64 is not defined and we cannot use:
            #   `(time[0] + time[-1]) / 2` directly
            epoch = time.values[0] + (time.values[-1] - time.values[0]) / 2

        if interval is None:
            interval = (time.values[0], time.values[-1])

        data = position.assign_coords({dim: time - epoch})
        polyfit_results = data.polyfit(dim=dim, deg=deg)
        # TODO: raise if the fit is not good enough

        return cls(polyfit_results.polyfit_coefficients, epoch, interval)

    def azimuth_time_range(self, freq_s: float = 0.02) -> xr.DataArray:
        azimuth_time_values = pd.date_range(
            start=self.interval[0],
            end=self.interval[-1],
            freq=pd.Timedelta(freq_s, "s"),
        ).values
        return xr.DataArray(
            azimuth_time_values,
            coords={"azimuth_time": azimuth_time_values},
            name="azimuth_time",
        )

    def position(
        self, time: Optional[xr.DataArray] = None, **kwargs: Any
    ) -> xr.DataArray:
        if time is None:
            time = self.azimuth_time_range(**kwargs)
        assert time.dtype.name in ("datetime64[ns]", "timedelta64[ns]")

        position: xr.DataArray
        position = xr.polyval(time - self.epoch, self.coefficients)
        position = position.assign_coords({time.name: time})
        return position.rename("position")

    def velocity(
        self, time: Optional[xr.DataArray] = None, **kwargs: Any
    ) -> xr.DataArray:
        if time is None:
            time = self.azimuth_time_range(**kwargs)
        assert time.dtype.name in ("datetime64[ns]", "timedelta64[ns]")

        velocity_coefficients = polyder(self.coefficients) * S_TO_NS

        velocity: xr.DataArray
        velocity = xr.polyval(time - self.epoch, velocity_coefficients)
        velocity = velocity.assign_coords({time.name: time})
        return velocity.rename("velocity")



import pathlib

import numpy as np
import pytest
import xarray as xr

from sarsen import sentinel1

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

DATA_PATHS = [
    DATA_FOLDER
    / "S1B_IW_GRDH_1SDV_20211223T051122_20211223T051147_030148_039993_5371.SAFE",
    DATA_FOLDER
    / "S1A_IW_SLC__1SDV_20220104T170557_20220104T170624_041314_04E951_F1F1.SAFE",
]

GROUPS = ["IW/VV", "IW1/VV"]


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
def test_Sentinel1SarProduct(data_path: str, group: str) -> None:
    res = sentinel1.Sentinel1SarProduct(data_path, group)

    assert isinstance(res.measurement, xr.Dataset)
    assert isinstance(res.orbit, xr.Dataset)
    assert isinstance(res.calibration, xr.Dataset)

    if res.product_type == "GRD":
        assert isinstance(res.coordinate_conversion, xr.Dataset)
        assert res.azimuth_fm_rate is None
        assert res.dc_estimate is None
    else:
        assert res.coordinate_conversion is None
        assert isinstance(res.azimuth_fm_rate, xr.Dataset)
        assert isinstance(res.dc_estimate, xr.Dataset)

    assert res.product_type in {"SLC", "GRD"}
    assert isinstance(res.beta_nought(), xr.DataArray)
    assert isinstance(res.state_vectors(), xr.DataArray)


def test_product_info() -> None:
    expected_geospatial_bbox = [
        11.86800305333565,
        40.87886713841886,
        15.32209672548896,
        42.78115380313222,
    ]
    product = sentinel1.Sentinel1SarProduct(str(DATA_PATHS[0]))

    res = product.product_info()

    assert "product_type" in res
    assert np.allclose(res["geospatial_bbox"], expected_geospatial_bbox)



import numpy as np
import xarray as xr

from sarsen import chunking


def test_compute_chunk_1d() -> None:
    ext_chunks, ext_chunks_bound, int_chunks = chunking.compute_chunks_1d(
        dim_size=20, chunks=10, bound=2
    )
    assert ext_chunks == [slice(0, 12), slice(8, 20)]
    assert int_chunks == [slice(0, 10), slice(10, 20)]
    assert ext_chunks_bound == [slice(0, 10), slice(2, 12)]

    ext_chunks, ext_chunks_bound, int_chunks = chunking.compute_chunks_1d(
        dim_size=13, chunks=10, bound=2
    )
    assert ext_chunks == [slice(0, 12), slice(8, 13)]
    assert int_chunks == [slice(0, 10), slice(10, 13)]
    assert ext_chunks_bound == [slice(0, 10), slice(2, 5)]

    # check bound case: k * chunks + bound == dim_size
    ext_chunks, ext_chunks_bound, int_chunks = chunking.compute_chunks_1d(
        dim_size=22, chunks=10, bound=2
    )
    assert ext_chunks == [slice(0, 12), slice(8, 22)]
    assert int_chunks == [slice(0, 10), slice(10, 22)]
    assert ext_chunks_bound == [slice(0, 10), slice(2, 14)]

    # check bound case: k * chunks + bound > dim_size >  k * chunks
    ext_chunks, ext_chunks_bound, int_chunks = chunking.compute_chunks_1d(
        dim_size=21, chunks=10, bound=2
    )
    assert ext_chunks == [slice(0, 12), slice(8, 21)]
    assert int_chunks == [slice(0, 10), slice(10, 21)]
    assert ext_chunks_bound == [slice(0, 10), slice(2, 13)]

    # check bound case: dim_size < bound
    ext_chunks, ext_chunks_bound, int_chunks = chunking.compute_chunks_1d(
        dim_size=2, chunks=10, bound=10
    )
    assert ext_chunks == [slice(0, 2)]
    assert int_chunks == [slice(0, 2)]
    assert ext_chunks_bound == [slice(0, 2)]


def test_compute_chunks() -> None:
    ext_chunks, ext_chunks_bound, int_chunks = chunking.compute_chunks(
        dims={"x": 10, "y": 21}, chunks=10, bound=2
    )
    assert {"x": slice(0, 10), "y": slice(0, 12)} in ext_chunks
    assert {"x": slice(0, 10), "y": slice(8, 21)} in ext_chunks

    assert {"x": slice(0, 10), "y": slice(0, 10)} in int_chunks
    assert {"x": slice(0, 10), "y": slice(10, 21)} in int_chunks

    assert {"x": slice(0, 10), "y": slice(0, 10)} in ext_chunks_bound
    assert {"x": slice(0, 10), "y": slice(2, 13)} in ext_chunks_bound


def test_map_ovelap() -> None:
    arr = xr.DataArray(np.arange(22 * 31).reshape((22, 31)), dims=("x", "y"))

    def function(x: xr.DataArray) -> xr.DataArray:
        return x

    res = chunking.map_ovelap(function=function, obj=arr, chunks=10, bound=2)
    assert res.equals(arr)



"""Reference "Guide to Sentinel-1 Geocoding" UZH-S1-GC-AD 1.10 26.03.2019.

See: https://sentinel.esa.int/documents/247904/0/Guide-to-Sentinel-1-Geocoding.pdf/e0450150-b4e9-4b2d-9b32-dadf989d3bd3
"""
import functools
from typing import Any, Callable, Optional, Tuple, TypeVar

import numpy as np
import numpy.typing as npt
import xarray as xr

TimedeltaArrayLike = TypeVar("TimedeltaArrayLike", bound=npt.ArrayLike)
FloatArrayLike = TypeVar("FloatArrayLike", bound=npt.ArrayLike)


def secant_method(
    ufunc: Callable[[TimedeltaArrayLike], Tuple[FloatArrayLike, FloatArrayLike]],
    t_prev: TimedeltaArrayLike,
    t_curr: TimedeltaArrayLike,
    diff_ufunc: float = 1.0,
    diff_t: np.timedelta64 = np.timedelta64(0, "ns"),
) -> Tuple[TimedeltaArrayLike, TimedeltaArrayLike, FloatArrayLike, Any]:
    """Return the root of ufunc calculated using the secant method."""
    # implementation modified from https://en.wikipedia.org/wiki/Secant_method
    f_prev, _ = ufunc(t_prev)

    # strong convergence, all points below one of the two thresholds
    while True:
        f_curr, payload_curr = ufunc(t_curr)

        # the `not np.any` construct let us accept `np.nan` as good values
        if not np.any((np.abs(f_curr) > diff_ufunc)):
            break

        t_diff: TimedeltaArrayLike
        p: TimedeltaArrayLike
        q: FloatArrayLike

        t_diff = t_curr - t_prev  # type: ignore
        p = f_curr * t_diff  # type: ignore
        q = f_curr - f_prev  # type: ignore

        # t_prev, t_curr = t_curr, t_curr - f_curr * np.timedelta64(-148_000, "ns")
        t_prev, t_curr = t_curr, t_curr - np.where(q != 0, p / q, 0)  # type: ignore
        f_prev = f_curr

        # the `not np.any` construct let us accept `np.nat` as good values
        if not np.any(np.abs(t_diff) > diff_t):
            break

    return t_curr, t_prev, f_curr, payload_curr


# FIXME: interpolationg the direction decreses the precision, this function should
#   probably have velocity_ecef_sar in input instead
def zero_doppler_plane_distance(
    dem_ecef: xr.DataArray,
    position_ecef_sar: xr.DataArray,
    direction_ecef_sar: xr.DataArray,
    azimuth_time: TimedeltaArrayLike,
    dim: str = "axis",
) -> Tuple[xr.DataArray, Tuple[xr.DataArray, xr.DataArray]]:
    dem_distance = dem_ecef - position_ecef_sar.interp(azimuth_time=azimuth_time)
    satellite_direction = direction_ecef_sar.interp(azimuth_time=azimuth_time)
    plane_distance = (dem_distance * satellite_direction).sum(dim, skipna=False)
    return plane_distance, (dem_distance, satellite_direction)


def backward_geocode(
    dem_ecef: xr.DataArray,
    position_ecef: xr.DataArray,
    velocity_ecef: xr.DataArray,
    azimuth_time: Optional[xr.DataArray] = None,
    dim: str = "axis",
    diff_ufunc: float = 1.0,
) -> xr.Dataset:
    direction_ecef = (
        velocity_ecef / xr.dot(velocity_ecef, velocity_ecef, dims=dim) ** 0.5
    )

    zero_doppler = functools.partial(
        zero_doppler_plane_distance, dem_ecef, position_ecef, direction_ecef
    )

    if azimuth_time is None:
        azimuth_time = position_ecef.azimuth_time
    t_template = dem_ecef.isel({dim: 0}).drop_vars(dim)
    t_prev = xr.full_like(t_template, azimuth_time.values[0], dtype=azimuth_time.dtype)
    t_curr = xr.full_like(t_template, azimuth_time.values[-1], dtype=azimuth_time.dtype)

    # NOTE: dem_distance has the associated azimuth_time as a coordinate already
    _, _, _, (dem_distance, satellite_direction) = secant_method(
        zero_doppler,
        t_prev,
        t_curr,
        diff_ufunc,
    )
    acquisition = xr.Dataset(
        data_vars={
            "dem_distance": dem_distance,
            "satellite_direction": satellite_direction.transpose(*dem_distance.dims),
        }
    )
    return acquisition.reset_coords("azimuth_time")



import os
import pathlib

import py
import pytest
import xarray as xr

from sarsen import apps, sentinel1

DATA_FOLDER = pathlib.Path(__file__).parent / "data"

DATA_PATHS = [
    DATA_FOLDER
    / "S1B_IW_GRDH_1SDV_20211223T051122_20211223T051147_030148_039993_5371.SAFE",
    DATA_FOLDER
    / "S1A_IW_SLC__1SDV_20220104T170557_20220104T170624_041314_04E951_F1F1.SAFE",
]

GROUPS = ["IW/VV", "IW1/VV"]

DEM_RASTER = DATA_FOLDER / "Rome-30m-DEM.tif"


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="too much memory")
def test_terrain_correction_gtc(
    tmpdir: py.path.local,
    data_path: pathlib.Path,
    group: str,
) -> None:
    out = str(tmpdir.join("GTC.tif"))
    product = sentinel1.Sentinel1SarProduct(
        str(data_path),
        group,
    )

    res = apps.terrain_correction(
        product,
        str(DEM_RASTER),
        output_urlpath=out,
    )

    assert isinstance(res, xr.DataArray)
    assert "beta" in res.attrs["long_name"]


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="too much memory")
def test_terrain_correction_fast_rtc(
    tmpdir: py.path.local, data_path: pathlib.Path, group: str
) -> None:
    out = str(tmpdir.join("RTC.tif"))
    product = sentinel1.Sentinel1SarProduct(
        str(data_path),
        group,
    )

    res = apps.terrain_correction(
        product,
        str(DEM_RASTER),
        correct_radiometry="gamma_nearest",
        output_urlpath=out,
    )

    assert isinstance(res, xr.DataArray)
    assert "gamma" in res.attrs["long_name"]


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="too much memory")
def test_terrain_correction_rtc(
    tmpdir: py.path.local, data_path: pathlib.Path, group: str
) -> None:
    out = str(tmpdir.join("RTC.tif"))
    product = sentinel1.Sentinel1SarProduct(
        str(data_path),
        group,
    )

    res = apps.terrain_correction(
        product,
        str(DEM_RASTER),
        correct_radiometry="gamma_bilinear",
        output_urlpath=out,
    )

    assert isinstance(res, xr.DataArray)
    assert "gamma" in res.attrs["long_name"]


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="too much memory")
def test_terrain_correction_gtc_dask(
    tmpdir: py.path.local, data_path: pathlib.Path, group: str
) -> None:
    out = str(tmpdir.join("GTC.tif"))
    product = sentinel1.Sentinel1SarProduct(
        str(data_path),
        group,
    )

    res = apps.terrain_correction(
        product,
        str(DEM_RASTER),
        output_urlpath=out,
        chunks=1024,
    )

    assert isinstance(res, xr.DataArray)
    assert "beta" in res.attrs["long_name"]


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="too much memory")
def test_terrain_correction_fast_rtc_dask(
    tmpdir: py.path.local, data_path: pathlib.Path, group: str
) -> None:
    out = str(tmpdir.join("RTC.tif"))
    product = sentinel1.Sentinel1SarProduct(
        str(data_path),
        group,
    )

    res = apps.terrain_correction(
        product,
        str(DEM_RASTER),
        correct_radiometry="gamma_nearest",
        output_urlpath=out,
        chunks=1024,
    )

    assert isinstance(res, xr.DataArray)
    assert "gamma" in res.attrs["long_name"]


@pytest.mark.parametrize("data_path,group", zip(DATA_PATHS, GROUPS))
@pytest.mark.skipif(os.getenv("GITHUB_ACTIONS") == "true", reason="too much memory")
def test_terrain_correction_rtc_dask(
    tmpdir: py.path.local, data_path: pathlib.Path, group: str
) -> None:
    out = str(tmpdir.join("RTC.tif"))
    product = sentinel1.Sentinel1SarProduct(
        str(data_path),
        group,
    )

    res = apps.terrain_correction(
        product,
        str(DEM_RASTER),
        correct_radiometry="gamma_bilinear",
        output_urlpath=out,
        chunks=1024,
    )

    assert isinstance(res, xr.DataArray)
    assert "gamma" in res.attrs["long_name"]



from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Union

import attrs
import numpy as np
import xarray as xr
import xarray_sentinel

import sarsen

try:
    import dask  # noqa: F401

    DEFAULT_MEASUREMENT_CHUNKS: Optional[int] = 2048
except ModuleNotFoundError:
    DEFAULT_MEASUREMENT_CHUNKS = None

SPEED_OF_LIGHT = 299_792_458.0  # m / s


def open_dataset_autodetect(
    product_urlpath: str,
    group: Optional[str] = None,
    chunks: Optional[Union[int, Dict[str, int]]] = None,
    check_files_exist: bool = False,
    **kwargs: Any,
) -> Tuple[xr.Dataset, Dict[str, Any]]:
    kwargs.setdefault("engine", "sentinel-1")
    try:
        ds = xr.open_dataset(
            product_urlpath,
            group=group,
            chunks=chunks,
            check_files_exist=check_files_exist,
            **kwargs,
        )
    except FileNotFoundError:
        # re-try with Planetary Computer option
        kwargs[
            "override_product_files"
        ] = "{dirname}/{prefix}{swath}-{polarization}{ext}"
        ds = xr.open_dataset(product_urlpath, group=group, chunks=chunks, **kwargs)
    return ds, kwargs


def azimuth_slant_range_grid(
    attrs: Dict[str, Any],
    grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
) -> Dict[str, Any]:
    if attrs["product_type"] == "SLC":
        slant_range_spacing_m = (
            attrs["range_pixel_spacing"]
            * np.sin(attrs["incidence_angle_mid_swath"])
            * grouping_area_factor[1]
        )
    else:
        slant_range_spacing_m = attrs["range_pixel_spacing"] * grouping_area_factor[1]

    slant_range_time_interval_s = (
        slant_range_spacing_m * 2 / SPEED_OF_LIGHT  # ignore type
    )

    grid_parameters: Dict[str, Any] = {
        "slant_range_time0": attrs["image_slant_range_time"],
        "slant_range_time_interval_s": slant_range_time_interval_s,
        "slant_range_spacing_m": slant_range_spacing_m,
        "azimuth_time0": np.datetime64(attrs["product_first_line_utc_time"]),
        "azimuth_time_interval_s": attrs["azimuth_time_interval"]
        * grouping_area_factor[0],
        "azimuth_spacing_m": attrs["azimuth_pixel_spacing"] * grouping_area_factor[0],
    }
    return grid_parameters


@attrs.define(slots=False)
class Sentinel1SarProduct(sarsen.GroundRangeSarProduct, sarsen.SlantRangeSarProduct):
    product_urlpath: str
    measurement_group: Optional[str] = None
    measurement_chunks: Optional[int] = DEFAULT_MEASUREMENT_CHUNKS
    kwargs: Dict[str, Any] = {}

    def all_measurement_groups(self) -> List[str]:
        ds, self.kwargs = open_dataset_autodetect(
            self.product_urlpath, check_files_exist=True, **self.kwargs
        )

        measurement_groups = [g for g in ds.attrs["subgroups"] if g.count("/") == 1]
        return measurement_groups

    @property
    def measurement(self) -> xr.Dataset:
        ds, self.kwargs = open_dataset_autodetect(
            self.product_urlpath,
            group=self.measurement_group,
            chunks=self.measurement_chunks,
            **self.kwargs,
        )
        if ds.attrs["product_type"] == "SLC" and ds.attrs["mode"] == "IW":
            ds = xarray_sentinel.mosaic_slc_iw(ds)
        return ds

    @property
    def orbit(self) -> xr.Dataset:
        ds, self.kwargs = open_dataset_autodetect(
            self.product_urlpath, group=f"{self.measurement_group}/orbit", **self.kwargs
        )
        return ds

    @property
    def gcp(self) -> xr.Dataset:
        ds, self.kwargs = open_dataset_autodetect(
            self.product_urlpath, group=f"{self.measurement_group}/gcp", **self.kwargs
        )
        return ds

    @property
    def calibration(self) -> xr.Dataset:
        ds, self.kwargs = open_dataset_autodetect(
            self.product_urlpath,
            group=f"{self.measurement_group}/calibration",
            **self.kwargs,
        )
        return ds

    @property
    def coordinate_conversion(self) -> Optional[xr.Dataset]:
        ds = None
        if self.product_type == "GRD":
            ds, self.kwargs = open_dataset_autodetect(
                self.product_urlpath,
                group=f"{self.measurement_group}/coordinate_conversion",
                **self.kwargs,
            )
        return ds

    @property
    def azimuth_fm_rate(self) -> Optional[xr.Dataset]:
        ds = None
        if self.product_type == "SLC":
            ds, self.kwargs = open_dataset_autodetect(
                self.product_urlpath,
                group=f"{self.measurement_group}/azimuth_fm_rate",
                **self.kwargs,
            )
        return ds

    @property
    def dc_estimate(self) -> Optional[xr.Dataset]:
        ds = None
        if self.product_type == "SLC":
            ds, self.kwargs = open_dataset_autodetect(
                self.product_urlpath,
                group=f"{self.measurement_group}/dc_estimate",
                **self.kwargs,
            )
        return ds

    # SarProduct interaface

    @property
    def product_type(self) -> Any:
        prod_type = self.measurement.attrs["product_type"]
        assert isinstance(prod_type, str)
        return prod_type

    def beta_nought(self) -> xr.DataArray:
        measurement = self.measurement.data_vars["measurement"]
        beta_nought = xarray_sentinel.calibrate_intensity(
            measurement, self.calibration.betaNought
        )
        beta_nought = beta_nought.drop_vars(["pixel", "line"])
        return beta_nought

    def state_vectors(self) -> xr.DataArray:
        return self.orbit.data_vars["position"]

    def slant_range_time_to_ground_range(
        self, azimuth_time: xr.DataArray, slant_range_time: xr.DataArray
    ) -> xr.DataArray:
        assert self.coordinate_conversion is not None
        # the dask graph explodes without the map_blocks due to the interpolations
        ground_range = xr.map_blocks(
            xarray_sentinel.slant_range_time_to_ground_range,
            azimuth_time,
            args=(slant_range_time,),
            kwargs={"coordinate_conversion": self.coordinate_conversion},
        )
        return ground_range

    def grid_parameters(
        self,
        grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
    ) -> Dict[str, Any]:
        return azimuth_slant_range_grid(self.measurement.attrs, grouping_area_factor)

    def complex_amplitude(self) -> xr.DataArray:
        measurement = self.measurement.data_vars["measurement"]
        beta_nought = xarray_sentinel.calibrate_amplitude(
            measurement, self.calibration.betaNought
        )
        beta_nought = beta_nought.drop_vars(["pixel", "line"])
        return beta_nought

    def interp_sar(self, *args: Any, **kwargs: Any) -> xr.DataArray:
        if self.product_type == "GRD":
            return sarsen.GroundRangeSarProduct.interp_sar(self, *args, **kwargs)
        else:
            return sarsen.SlantRangeSarProduct.interp_sar(self, *args, **kwargs)

    def product_info(self, **kwargs: Any) -> Dict[str, Any]:
        """Get information about the Sentinel-1 product."""
        measurement_groups = self.all_measurement_groups()

        self.measurement_group = self.measurement_group or measurement_groups[0]
        gcp = self.gcp

        bbox = [
            gcp.attrs["geospatial_lon_min"],
            gcp.attrs["geospatial_lat_min"],
            gcp.attrs["geospatial_lon_max"],
            gcp.attrs["geospatial_lat_max"],
        ]

        product_attrs = [
            "product_type",
            "mode",
            "swaths",
            "transmitter_receiver_polarisations",
            "relative_orbit_number",
            "orbit_number",
            "mission_data_take_id",
        ]
        product_info = {}
        for attr_name in product_attrs:
            try:
                product_info[attr_name] = gcp.attrs[attr_name]
            except KeyError:
                pass
        product_info.update(
            {
                "measurement_groups": measurement_groups,
                "geospatial_bounds": gcp.attrs["geospatial_bounds"],
                "geospatial_bbox": bbox,
            }
        )

        return product_info



import json
import logging
from typing import Tuple

import typer

from . import apps, sentinel1

app = typer.Typer()


@app.command()
def info(
    product_urlpath: str,
) -> None:
    """Print information about the Sentinel-1 product."""
    logging.basicConfig(level=logging.INFO)
    product = sentinel1.Sentinel1SarProduct(product_urlpath)
    product_info = product.product_info()
    for key, value in product_info.items():
        print(f"{key}: {value}")


@app.command()
def gtc(
    product_urlpath: str,
    measurement_group: str,
    dem_urlpath: str,
    output_urlpath: str = "GTC.tif",
    enable_dask_distributed: bool = False,
    client_kwargs_json: str = '{"processes": false}',
    chunks: int = 1024,
) -> None:
    """Generate a geometrically terrain corrected (GTC) image from Sentinel-1 product."""
    client_kwargs = json.loads(client_kwargs_json)
    real_chunks = chunks if chunks > 0 else None
    logging.basicConfig(level=logging.INFO)
    product = sentinel1.Sentinel1SarProduct(
        product_urlpath,
        measurement_group,
    )
    apps.terrain_correction(
        product,
        dem_urlpath,
        output_urlpath=output_urlpath,
        enable_dask_distributed=enable_dask_distributed,
        client_kwargs=client_kwargs,
        chunks=real_chunks,
    )


@app.command()
def stc(
    product_urlpath: str,
    measurement_group: str,
    dem_urlpath: str,
    simulated_urlpath: str = "STC.tif",
    enable_dask_distributed: bool = False,
    client_kwargs_json: str = '{"processes": false}',
    chunks: int = 1024,
    grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
) -> None:
    """Generate a simulated terrain corrected image from a Sentinel-1 product."""
    client_kwargs = json.loads(client_kwargs_json)
    real_chunks = chunks if chunks > 0 else None
    logging.basicConfig(level=logging.INFO)
    product = sentinel1.Sentinel1SarProduct(
        product_urlpath,
        measurement_group,
    )
    apps.terrain_correction(
        product,
        dem_urlpath,
        correct_radiometry="gamma_bilinear",
        simulated_urlpath=simulated_urlpath,
        grouping_area_factor=grouping_area_factor,
        enable_dask_distributed=enable_dask_distributed,
        client_kwargs=client_kwargs,
        chunks=real_chunks,
    )


@app.command()
def rtc(
    product_urlpath: str,
    measurement_group: str,
    dem_urlpath: str,
    output_urlpath: str = "RTC.tif",
    enable_dask_distributed: bool = False,
    client_kwargs_json: str = '{"processes": false}',
    chunks: int = 1024,
    grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
) -> None:
    """Generate a radiometrically terrain corrected (RTC) image from Sentinel-1 product."""
    client_kwargs = json.loads(client_kwargs_json)
    real_chunks = chunks if chunks > 0 else None
    logging.basicConfig(level=logging.INFO)
    product = sentinel1.Sentinel1SarProduct(
        product_urlpath,
        measurement_group,
    )
    apps.terrain_correction(
        product,
        dem_urlpath,
        correct_radiometry="gamma_bilinear",
        output_urlpath=output_urlpath,
        grouping_area_factor=grouping_area_factor,
        enable_dask_distributed=enable_dask_distributed,
        client_kwargs=client_kwargs,
        chunks=real_chunks,
    )


if __name__ == "__main__":
    app()



import logging
from typing import Any, Container, Dict, Optional, Tuple
from unittest import mock

import numpy as np
import rioxarray
import xarray as xr

from . import chunking, datamodel, geocoding, orbit, radiometry, scene

logger = logging.getLogger(__name__)


SPEED_OF_LIGHT = 299_792_458.0  # m / s


def simulate_acquisition(
    dem_ecef: xr.DataArray,
    position_ecef: xr.DataArray,
    include_variables: Container[str] = (),
) -> xr.Dataset:
    """Compute the image coordinates of the DEM given the satellite orbit."""
    orbit_interpolator = orbit.OrbitPolyfitIterpolator.from_position(position_ecef)
    position_ecef = orbit_interpolator.position()
    velocity_ecef = orbit_interpolator.velocity()

    acquisition = geocoding.backward_geocode(dem_ecef, position_ecef, velocity_ecef)

    slant_range = (acquisition.dem_distance**2).sum(dim="axis") ** 0.5
    slant_range_time = 2.0 / SPEED_OF_LIGHT * slant_range

    acquisition["slant_range_time"] = slant_range_time

    if include_variables and "gamma_area" in include_variables:
        gamma_area = radiometry.compute_gamma_area(
            dem_ecef, acquisition.dem_distance / slant_range
        )
        acquisition["gamma_area"] = gamma_area

    for data_var_name in acquisition.data_vars:
        if include_variables and data_var_name not in include_variables:
            acquisition = acquisition.drop_vars(data_var_name)  # type: ignore

    # drop coordinates that are not associated with any data variable
    for coord_name in acquisition.coords:
        if all(coord_name not in dv.coords for dv in acquisition.data_vars.values()):
            acquisition = acquisition.drop_vars(coord_name)  # type: ignore

    return acquisition


def map_simulate_acquisition(
    dem_ecef: xr.DataArray,
    position_ecef: xr.DataArray,
    template_raster: xr.DataArray,
    correct_radiometry: Optional[str] = None,
) -> xr.Dataset:
    acquisition_template = xr.Dataset(
        data_vars={
            "slant_range_time": template_raster,
            "azimuth_time": template_raster.astype("datetime64[ns]"),
        }
    )
    include_variables = {"slant_range_time", "azimuth_time"}
    if correct_radiometry is not None:
        acquisition_template["gamma_area"] = template_raster
        include_variables.add("gamma_area")

    acquisition = xr.map_blocks(
        simulate_acquisition,
        dem_ecef,
        kwargs={
            "position_ecef": position_ecef,
            "include_variables": include_variables,
        },
        template=acquisition_template,
    )
    return acquisition


def do_terrain_correction(
    product: datamodel.SarProduct,
    dem_raster: xr.DataArray,
    correct_radiometry: Optional[str] = None,
    interp_method: xr.core.types.InterpOptions = "nearest",
    grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
    radiometry_chunks: int = 2048,
    radiometry_bound: int = 128,
) -> tuple[xr.DataArray, Optional[xr.DataArray]]:
    logger.info("pre-process DEM")

    dem_ecef = xr.map_blocks(
        scene.convert_to_dem_ecef, dem_raster, kwargs={"source_crs": dem_raster.rio.crs}
    )
    dem_ecef = dem_ecef.drop_vars(dem_ecef.rio.grid_mapping)

    logger.info("simulate acquisition")

    template_raster = dem_raster.drop_vars(dem_raster.rio.grid_mapping) * 0.0

    acquisition = map_simulate_acquisition(
        dem_ecef, product.state_vectors(), template_raster, correct_radiometry
    )

    simulated_beta_nought = None
    if correct_radiometry is not None:
        logger.info("simulate radiometry")

        grid_parameters = product.grid_parameters(grouping_area_factor)

        if correct_radiometry == "gamma_bilinear":
            gamma_weights = radiometry.gamma_weights_bilinear
        elif correct_radiometry == "gamma_nearest":
            gamma_weights = radiometry.gamma_weights_nearest

        acquisition = acquisition.persist()

        simulated_beta_nought = chunking.map_ovelap(
            obj=acquisition,
            function=gamma_weights,
            chunks=radiometry_chunks,
            bound=radiometry_bound,
            kwargs=grid_parameters,
            template=template_raster,
        )
        simulated_beta_nought.attrs["long_name"] = "terrain-simulated beta nought"

        simulated_beta_nought.x.attrs.update(dem_raster.x.attrs)
        simulated_beta_nought.y.attrs.update(dem_raster.y.attrs)
        simulated_beta_nought.rio.write_crs(dem_raster.rio.crs, inplace=True)

    logger.info("calibrate image")

    beta_nought = product.beta_nought()

    logger.info("terrain-correct image")

    # HACK: we monkey-patch away an optimisation in xr.DataArray.interp that actually makes
    #   the interpolation much slower when indeces are dask arrays.
    with mock.patch("xarray.core.missing._localize", lambda o, i: (o, i)):
        geocoded = product.interp_sar(
            beta_nought,
            azimuth_time=acquisition.azimuth_time,
            slant_range_time=acquisition.slant_range_time,
            method=interp_method,
        )

    if correct_radiometry is not None:
        assert simulated_beta_nought is not None
        geocoded = geocoded / simulated_beta_nought
        geocoded.attrs["long_name"] = "terrain-corrected gamma nought"

    geocoded.x.attrs.update(dem_raster.x.attrs)
    geocoded.y.attrs.update(dem_raster.y.attrs)
    geocoded.rio.write_crs(dem_raster.rio.crs, inplace=True)

    return geocoded, simulated_beta_nought


def terrain_correction(
    product: datamodel.SarProduct,
    dem_urlpath: str,
    output_urlpath: Optional[str] = "GTC.tif",
    simulated_urlpath: Optional[str] = None,
    correct_radiometry: Optional[str] = None,
    interp_method: xr.core.types.InterpOptions = "nearest",
    grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
    open_dem_raster_kwargs: Dict[str, Any] = {},
    chunks: Optional[int] = 1024,
    radiometry_chunks: int = 2048,
    radiometry_bound: int = 128,
    enable_dask_distributed: bool = False,
    client_kwargs: Dict[str, Any] = {"processes": False},
) -> xr.DataArray:
    """Apply the terrain-correction to sentinel-1 SLC and GRD products.

    :param product: SarProduct instance representing the input data
    :param dem_urlpath: dem path or url
    :param output_urlpath: output path or url
    :param correct_radiometry: default `None`. If `correct_radiometry=None`the radiometric terrain
    correction is not applied. `correct_radiometry=gamma_bilinear` applies the gamma flattening classic
    algorithm using bilinear interpolation to compute the weights. `correct_radiometry=gamma_nearest`
    applies the gamma flattening using nearest neighbours instead of bilinear interpolation.
    'gamma_nearest' significantly reduces the processing time
    :param interp_method: interpolation method for product resampling.
    The interpolation methods are the methods supported by ``xarray.DataArray.interp``
    :param grouping_area_factor: is a tuple of floats greater than 1. The default is `(1, 1)`.
    The `grouping_area_factor`  can be increased (i) to speed up the processing or
    (ii) when the input DEM resolution is low.
    The Gamma Flattening usually works properly if the pixel size of the input DEM is much smaller
    than the pixel size of the input Sentinel-1 product.
    Otherwise, the output may have radiometric distortions.
    This problem can be avoided by increasing the `grouping_area_factor`.
    Be aware that `grouping_area_factor` too high may degrade the final result
    :param open_dem_raster_kwargs: additional keyword arguments passed on to ``xarray.open_dataset``
    to open the `dem_urlpath`
    """
    # rioxarray must be imported explicitly or accesses to `.rio` may fail in dask
    assert rioxarray.__version__

    allowed_correct_radiometry = [None, "gamma_bilinear", "gamma_nearest"]
    if correct_radiometry not in allowed_correct_radiometry:
        raise ValueError(
            f"{correct_radiometry=}. Must be one of: {allowed_correct_radiometry}"
        )
    if simulated_urlpath is not None and correct_radiometry is None:
        raise ValueError("Simulation cannot be saved")
    if output_urlpath is None and simulated_urlpath is None:
        raise ValueError("No output selected")

    output_chunks = chunks if chunks is not None else 512

    to_raster_kwargs: Dict[str, Any] = {}
    if enable_dask_distributed:
        from dask.distributed import Client, Lock

        client = Client(**client_kwargs)  # type: ignore
        to_raster_kwargs["lock"] = Lock("rio", client=client)  # type: ignore
        to_raster_kwargs["compute"] = False
        print(f"Dask distributed dashboard at: {client.dashboard_link}")

    logger.info(f"open DEM {dem_urlpath!r}")

    dem_raster = scene.open_dem_raster(
        dem_urlpath, chunks=chunks, **open_dem_raster_kwargs
    )

    allowed_product_types = ["GRD", "SLC"]
    if product.product_type not in allowed_product_types:
        raise ValueError(
            f"{product.product_type=}. Must be one of: {allowed_product_types}"
        )

    geocoded, simulated_beta_nought = do_terrain_correction(
        product=product,
        dem_raster=dem_raster,
        correct_radiometry=correct_radiometry,
        interp_method=interp_method,
        grouping_area_factor=grouping_area_factor,
        radiometry_chunks=radiometry_chunks,
        radiometry_bound=radiometry_bound,
    )

    if simulated_urlpath is not None:
        assert simulated_beta_nought is not None
        if output_urlpath is not None:
            simulated_beta_nought.persist()

        logger.info("save simulated")

        maybe_delayed = simulated_beta_nought.rio.to_raster(
            simulated_urlpath,
            dtype=np.float32,
            tiled=True,
            blockxsize=output_chunks,
            blockysize=output_chunks,
            compress="ZSTD",
            num_threads="ALL_CPUS",
            **to_raster_kwargs,
        )

        if enable_dask_distributed:
            maybe_delayed.compute()

    if output_urlpath is None:
        assert simulated_beta_nought is not None
        return simulated_beta_nought

    logger.info("save output")

    maybe_delayed = geocoded.rio.to_raster(
        output_urlpath,
        dtype=np.float32,
        tiled=True,
        blockxsize=output_chunks,
        blockysize=output_chunks,
        compress="ZSTD",
        num_threads="ALL_CPUS",
        **to_raster_kwargs,
    )

    if enable_dask_distributed:
        maybe_delayed.compute()

    return geocoded



import abc
from typing import Any, Dict, Optional, Tuple

import xarray as xr


class SarProduct(abc.ABC):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    @property
    @abc.abstractmethod
    def product_type(self) -> str:
        ...

    @abc.abstractmethod
    def state_vectors(self) -> xr.DataArray:
        ...

    @abc.abstractmethod
    def beta_nought(self) -> xr.DataArray:
        ...

    @abc.abstractmethod
    def interp_sar(
        self,
        data: xr.DataArray,
        azimuth_time: xr.DataArray,
        slant_range_time: Optional[xr.DataArray] = None,
        method: xr.core.types.InterpOptions = "nearest",
        ground_range: Optional[xr.DataArray] = None,
    ) -> xr.DataArray:
        ...

    # FIXME: design a better interface
    @abc.abstractmethod
    def grid_parameters(
        self,
        grouping_area_factor: Tuple[float, float] = (3.0, 3.0),
    ) -> Dict[str, Any]:
        ...


class GroundRangeSarProduct(SarProduct):
    @abc.abstractmethod
    def slant_range_time_to_ground_range(
        self, azimuth_time: xr.DataArray, slant_range_time: xr.DataArray
    ) -> xr.DataArray:
        ...

    def interp_sar(
        self,
        data: xr.DataArray,
        azimuth_time: xr.DataArray,
        slant_range_time: Optional[xr.DataArray] = None,
        method: xr.core.types.InterpOptions = "nearest",
        ground_range: Optional[xr.DataArray] = None,
    ) -> xr.DataArray:
        if ground_range is None:
            assert slant_range_time is not None
            ground_range = self.slant_range_time_to_ground_range(
                azimuth_time, slant_range_time
            )
        interpolated = data.interp(
            azimuth_time=azimuth_time, ground_range=ground_range, method=method
        )
        return interpolated.assign_attrs(data.attrs)


class SlantRangeSarProduct(SarProduct):
    @abc.abstractmethod
    def complex_amplitude(self) -> xr.DataArray:
        ...

    def beta_nought(self) -> xr.DataArray:
        amplitude = self.complex_amplitude()
        assert amplitude.attrs["units"] == "m m-1"
        beta_nought = abs(amplitude) ** 2
        return beta_nought.assign_attrs(long_name="beta nought", units="m2 m-2")

    def interp_sar(
        self,
        data: xr.DataArray,
        azimuth_time: xr.DataArray,
        slant_range_time: Optional[xr.DataArray] = None,
        method: xr.core.types.InterpOptions = "nearest",
        ground_range: Optional[xr.DataArray] = None,
    ) -> xr.DataArray:
        assert ground_range is None
        interpolated = data.interp(
            azimuth_time=azimuth_time, slant_range_time=slant_range_time, method=method
        )
        return interpolated.assign_attrs(data.attrs)



import logging
from typing import Optional, Tuple
from unittest import mock

import flox.xarray
import numpy as np
import xarray as xr

from . import scene

logger = logging.getLogger(__name__)

ONE_SECOND = np.timedelta64(1, "s")


def sum_weights(
    initial_weights: xr.DataArray,
    azimuth_index: xr.DataArray,
    slant_range_index: xr.DataArray,
    multilook: Optional[Tuple[int, int]] = None,
) -> xr.DataArray:
    geocoded = initial_weights.assign_coords(
        slant_range_index=slant_range_index, azimuth_index=azimuth_index
    )

    flat_sum: xr.DataArray = flox.xarray.xarray_reduce(
        geocoded,
        geocoded.slant_range_index,
        geocoded.azimuth_index,
        func="sum",
        method="map-reduce",
    )

    if multilook:
        flat_sum = flat_sum.rolling(
            azimuth_index=multilook[0],
            slant_range_index=multilook[1],
            center=True,
            min_periods=multilook[0] * multilook[1] // 2 + 1,
        ).mean()

    with mock.patch("xarray.core.missing._localize", lambda o, i: (o, i)):
        weights_sum = flat_sum.interp(
            slant_range_index=slant_range_index,
            azimuth_index=azimuth_index,
            method="nearest",
        )

    return weights_sum


def compute_gamma_area(
    dem_ecef: xr.DataArray,
    dem_direction: xr.DataArray,
) -> xr.DataArray:
    dem_oriented_area = scene.compute_dem_oriented_area(dem_ecef)
    gamma_area: xr.DataArray = xr.dot(dem_oriented_area, -dem_direction, dims="axis")
    gamma_area = gamma_area.where(gamma_area > 0, 0)
    return gamma_area


def gamma_weights_bilinear(
    dem_coords: xr.Dataset,
    slant_range_time0: float,
    azimuth_time0: np.datetime64,
    slant_range_time_interval_s: float,
    azimuth_time_interval_s: float,
    slant_range_spacing_m: float = 1.0,
    azimuth_spacing_m: float = 1.0,
) -> xr.DataArray:
    # compute dem image coordinates
    azimuth_index = ((dem_coords.azimuth_time - azimuth_time0) / ONE_SECOND) / (
        azimuth_time_interval_s
    )

    slant_range_index = (dem_coords.slant_range_time - slant_range_time0) / (
        slant_range_time_interval_s
    )

    slant_range_index_0 = np.floor(slant_range_index).astype(int).compute()
    slant_range_index_1 = np.ceil(slant_range_index).astype(int).compute()
    azimuth_index_0 = np.floor(azimuth_index).astype(int).compute()
    azimuth_index_1 = np.ceil(azimuth_index).astype(int).compute()

    logger.info("compute gamma areas 1/4")
    w_00 = abs(
        (azimuth_index_1 - azimuth_index) * (slant_range_index_1 - slant_range_index)
    )
    tot_area_00 = sum_weights(
        dem_coords["gamma_area"] * w_00,
        azimuth_index=azimuth_index_0,
        slant_range_index=slant_range_index_0,
    )

    logger.info("compute gamma areas 2/4")
    w_01 = abs(
        (azimuth_index_1 - azimuth_index) * (slant_range_index_0 - slant_range_index)
    )
    tot_area_01 = sum_weights(
        dem_coords["gamma_area"] * w_01,
        azimuth_index=azimuth_index_0,
        slant_range_index=slant_range_index_1,
    )

    logger.info("compute gamma areas 3/4")
    w_10 = abs(
        (azimuth_index_0 - azimuth_index) * (slant_range_index_1 - slant_range_index)
    )
    tot_area_10 = sum_weights(
        dem_coords["gamma_area"] * w_10,
        azimuth_index=azimuth_index_1,
        slant_range_index=slant_range_index_0,
    )

    logger.info("compute gamma areas 4/4")
    w_11 = abs(
        (azimuth_index_0 - azimuth_index) * (slant_range_index_0 - slant_range_index)
    )
    tot_area_11 = sum_weights(
        dem_coords["gamma_area"] * w_11,
        azimuth_index=azimuth_index_1,
        slant_range_index=slant_range_index_1,
    )

    tot_area = tot_area_00 + tot_area_01 + tot_area_10 + tot_area_11

    normalized_area = tot_area / (azimuth_spacing_m * slant_range_spacing_m)
    return normalized_area


def gamma_weights_nearest(
    dem_coords: xr.Dataset,
    slant_range_time0: float,
    azimuth_time0: np.datetime64,
    slant_range_time_interval_s: float,
    azimuth_time_interval_s: float,
    slant_range_spacing_m: float = 1.0,
    azimuth_spacing_m: float = 1.0,
) -> xr.DataArray:
    # compute dem image coordinates
    azimuth_index = np.round(
        (dem_coords.azimuth_time - azimuth_time0) / ONE_SECOND / azimuth_time_interval_s
    ).astype(int)

    slant_range_index = np.round(
        (dem_coords.slant_range_time - slant_range_time0) / slant_range_time_interval_s
    ).astype(int)

    logger.info("compute gamma areas 1/1")

    tot_area = sum_weights(
        dem_coords["gamma_area"],
        azimuth_index=azimuth_index,
        slant_range_index=slant_range_index,
    )

    normalized_area = tot_area / (azimuth_spacing_m * slant_range_spacing_m)
    return normalized_area



import numpy as np
import xarray as xr

from sarsen import orbit


def test_OrbitPolyfitIterpolator_datetime64(orbit_ds: xr.Dataset) -> None:
    position = orbit_ds.data_vars["position"]
    orbit_interpolator = orbit.OrbitPolyfitIterpolator.from_position(position, deg=4)

    res = orbit_interpolator.position(position.azimuth_time)

    assert res.dims == ("azimuth_time", "axis")
    assert np.allclose(res, position, rtol=0, atol=0.001)

    expected_velocity = orbit_ds.data_vars["velocity"]
    res = orbit_interpolator.velocity(position.azimuth_time)

    assert res.dims == ("azimuth_time", "axis")
    assert np.allclose(res, expected_velocity, rtol=0, atol=0.02)

    res = orbit_interpolator.position()
    assert res.dims == ("azimuth_time", "axis")

    res = orbit_interpolator.velocity()
    assert res.dims == ("azimuth_time", "axis")


def test_OrbitPolyfitIterpolator_timedelta64(orbit_ds: xr.Dataset) -> None:
    position = orbit_ds.data_vars["position"]
    position = position.assign_coords(
        azimuth_time=position.azimuth_time - position.azimuth_time[0]
    )
    epoch = position.azimuth_time.values[0]
    orbit_interpolator = orbit.OrbitPolyfitIterpolator.from_position(
        position, epoch=epoch, deg=4
    )

    res = orbit_interpolator.position(position.azimuth_time)

    assert res.dims == ("azimuth_time", "axis")
    assert np.allclose(res, position, rtol=0, atol=0.001)

    expected_velocity = orbit_ds.data_vars["velocity"]
    res = orbit_interpolator.velocity(position.azimuth_time)

    assert res.dims == ("azimuth_time", "axis")
    assert np.allclose(res, expected_velocity, rtol=0, atol=0.02)



import itertools
import math
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import xarray as xr


def compute_chunks_1d(
    dim_size: int,
    chunks: int = 2048,
    bound: int = 128,
) -> Tuple[List[slice], List[slice], List[slice]]:
    ext_slices = []
    ext_slices_bound = []
    int_slices = []

    # -bound is needed to avoid to incorporate the last chunk, if smaller of bound in the previous chunk
    if dim_size > bound:
        number_of_chunks = int(math.ceil((dim_size - bound) / chunks))
    else:
        number_of_chunks = 1
    for n in range(number_of_chunks):
        l_int = n * chunks
        if n * chunks - bound > 0:
            l_ext = n * chunks - bound
        else:
            l_ext = 0
        l_bound = l_int - l_ext

        if (n + 1) * chunks + bound < dim_size:
            r_ext = (n + 1) * chunks + bound
            r_int = (n + 1) * chunks
            r_bound = chunks + l_bound
        else:
            r_ext = dim_size
            r_int = dim_size
            r_bound = r_ext - l_ext

        ext_slices.append(slice(l_ext, r_ext))
        ext_slices_bound.append(slice(l_bound, r_bound))
        int_slices.append(slice(l_int, r_int))
    return ext_slices, ext_slices_bound, int_slices


def compute_product(
    slices: List[List[slice]], dims_name: List[str]
) -> List[Dict[str, slice]]:
    product: List[Dict[str, slice]] = []

    for slices_ in itertools.product(*slices):
        product.append({})
        for dim, sl in zip(dims_name, slices_):
            product[-1][dim] = sl
    return product


def compute_chunks(
    dims: Dict[str, int] = {},
    chunks: int = 2048,
    bound: int = 128,
) -> Tuple[List[Dict[str, slice]], List[Dict[str, slice]], List[Dict[str, slice]]]:
    ext_slices_ = []
    ext_slices_bound_ = []
    int_slices_ = []
    for dim_size in dims.values():
        ec, ecb, ic = compute_chunks_1d(dim_size, chunks=chunks, bound=bound)
        ext_slices_.append(ec)
        ext_slices_bound_.append(ecb)
        int_slices_.append(ic)

    ext_slices = compute_product(ext_slices_, list(dims))
    ext_slices_bound = compute_product(ext_slices_bound_, list(dims))
    int_slices = compute_product(int_slices_, list(dims))
    return ext_slices, ext_slices_bound, int_slices


def map_ovelap(
    function: Callable[..., xr.DataArray],
    obj: Union[xr.Dataset, xr.DataArray],
    chunks: int = 2048,
    bound: int = 128,
    kwargs: Dict[Any, Any] = {},
    template: Optional[xr.DataArray] = None,
) -> xr.DataArray:
    dims = {}
    for d in obj.dims:
        dims[str(d)] = len(obj[d])

    if isinstance(obj, xr.Dataset):
        if template is None:
            raise ValueError(
                "template argument is mandatory if obj is type of xr.Dataset"
            )
    elif isinstance(obj, xr.DataArray):
        if template is None:
            template = obj

    ext_chunks, ext_chunks_bounds, int_chunks = compute_chunks(
        dims, chunks, bound
    )  # type ignore

    try:
        from dask.array import empty_like  # type: ignore
    except ModuleNotFoundError:
        from numpy import empty_like  # type: ignore

    out = xr.DataArray(empty_like(template.data), dims=template.dims)  # type: ignore
    out.coords.update(obj.coords)
    for ext_chunk, ext_chunk_bounds, int_chunk in zip(
        ext_chunks, ext_chunks_bounds, int_chunks
    ):
        out_chunk = function(obj.isel(ext_chunk), **kwargs)
        out[int_chunk] = out_chunk.isel(ext_chunk_bounds)
    return out
